"use strict";


const fetch = require("node-fetch");

const junitXml = require("junit-xml");
var fs = require('fs');


const { URLSearchParams } = require('url');
const https = require("https");
const agent = new https.Agent({
  rejectUnauthorized: false
});

process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
var config = {};

console.log("Accelq AzureDevops Plugin");

const { Command } = require('commander');
const program = new Command();

program
  .usage('[OPTIONS]...')
  .option('-url, --url <name>', 'accelq portal url')
  .option('-t, --tenant <name>', 'Tenant name')
  .option('-p, --project <name>', 'Accelq Project name')
  .option('-u, --user <name>', 'Accelq Login User name')
  .option('-pwd, --password <passwd>', 'Accelq login password')
  .option('-jid, --jobid <name>', 'Accelq CICD job id')
  .option('-a, --agent <name>', 'Accelq agent name or pattern. use any for any available')
  .option('-rp, --runparams <name>', 'Accelq Runtime Params seperated by ambersand')
  .parse(process.argv);

const commander = program.opts();

console.log(commander);

var loadConfig = () => {
  return {
    url: commander.url,
    tenant: commander.tenant,
    project: commander.project,
    username:commander.user,
    // password:  new Buffer(commander.password).toString('base64'),
    password: commander.password,
    jobId: commander.jobid,
    agent: commander.agent,
    runParamPayload: commander.rp
  };
}


var JUNIT_REPORT = {};

console.log(loadConfig);

var ACCESS_TOKEN, CLIENT_ID, REFRESH_TOKEN, TENANT_CODE, BASE_URL, API_ENDPOINT, LOGIN_URL, USERS_URL,
    JOB_TRIGGER_URL, JOB_SUMMARY_URL, JOB_CONSIDATED_URL, AGENTS_URL;

var API_VERSION = "v1";

const doLogin = async () => {
  console.log("Password",config.password);
  try {
    var options = {
      method: 'POST',
      headers:{ 'cache-control': 'no-cache',
         'invalidate-active-tokens': 'true',
         'content-type': 'application/x-www-form-urlencoded'
      },
      agent:agent,
      body: new URLSearchParams({ username: config.username, password: config.password})
    };
    const response = await fetch(LOGIN_URL, options);
    if (response.status == 200) {
      info("Login successful!!!!");
      let json = await response.json();
      setSecurityTokens(json);
      return json;
    } else {
      console.error("Login Failed", response.status, response);
      return;
    }
  } catch (e) {
    console.error("ERROR", e);
  }
}

const generateJUnitReport = (consolidatedReportObj)=>{
  const JUNIT_REPORT = {};

  JUNIT_REPORT["name"] = consolidatedReportObj["purpose"];
  JUNIT_REPORT["suites"] = [];
  JUNIT_REPORT["suites"].push({});
  JUNIT_REPORT["suites"][0].name = consolidatedReportObj["purpose"];
  JUNIT_REPORT["suites"][0].timestamp = new Date(consolidatedReportObj["startTimestamp"]);

  JUNIT_REPORT["suites"][0].testCases = [];
  var totalTime = 0;
  var tcList = consolidatedReportObj["testCaseSummaryList"];
  console.log(tcList);
  if(tcList.length > 0){
    JUNIT_REPORT["suites"][0].name = tcList[0]["extResultLink"];
  }

  for(let i = 0; i < tcList.length; i++){
    let tc = {
      name: tcList[i]["name"]
    };
    if(tcList[i].status != "pass"){
      tc["failures"] = [{
         message: 'Test Case Failed. Please check the full accelq report for error details'
      }];
    }
    var timeDiff = (tcList[i]["endDate"] - tcList[i]["startDate"])/ 1000;
    totalTime += timeDiff;
    tc.time = timeDiff;
    tc["systemOut"] = [tcList[i]["extResultLink"]];
    JUNIT_REPORT["suites"][0].testCases.push(tc);
  }
  JUNIT_REPORT["suites"][0].time = totalTime;
  const xml = junitXml.getJunitXml(JUNIT_REPORT, {"schema":"ant-junit"});
  console.log(xml);
  fs.writeFileSync('accelq_report.xml', xml);
}
const setSecurityTokens = (loginResponse) => {
  ACCESS_TOKEN = loginResponse["access_token"];
  CLIENT_ID = loginResponse["client_id"];
  REFRESH_TOKEN = loginResponse["refresh_token"];
  TENANT_CODE = loginResponse["tenantCode"];
  var PROJECT_NAME;
  var userProjectList = loginResponse["userProjects"];

  for(var i = 0; i < userProjectList.length; i++){
    var userProject = userProjectList[i];
    //console.log(userProject);
    var projectName_ = userProject["projectDisplayName"];
    if(projectName_.toLowerCase().trim() == config["project"].toLowerCase().trim()) {
        PROJECT_NAME = userProject["projectName"];
        break;
    }
  }
  if(!PROJECT_NAME)
      return false;
  return true;
}

const getRunParamJsonPayload = (runParamStr) => {
    if(!runParamStr || runParamStr.trim().length == 0)
        return {};

    var runParamPayload = {};
    var splitOnAmp = runParamStr.split("&");
    splitOnAmp.forEach((paramInfo) => {
      var splitOnEquals = paramInfo.split("=");
      if(splitOnEquals.length == 2) {
        var key = splitOnEquals[0].trim(), value = splitOnEquals[1].trim();
        if(!(key == "") && !(value == "")) {
            runParamPayload[key] = value;
        }
      }
    });
    console.log("RUN PAYLOAD", runParamPayload);
    return runParamPayload;
}

const triggerJob = async (jobId, runParamPayload) => {
  console.log("triggering job:", jobId, " with ", runParamPayload);

  JOB_TRIGGER_URL = API_ENDPOINT + API_VERSION + "/test-exec/jobs/auto/" + jobId + "/trigger-ci";
  console.log("JOB TRIGGER URL", JOB_TRIGGER_URL);
  var options = {
    method: 'PUT',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN,
      "Content-Type": "application/json"
    },
    agent:agent,
    body:JSON.stringify(runParamPayload)
  };

  //console.log("options", options);
  //config.url + '/awb/api/aq_global/security/login'
  const response = await fetch(JOB_TRIGGER_URL, options);
  if (response.status == 200) {
    let json = await response.json();
    return json;
  } else {
    console.error("Failed to build job", jobId, " Status:", response.status, ":", response.statusText);
    return;
  }
}

const getUsers = async () => {
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  //console.log("options", options);
  //config.url + '/awb/api/aq_global/security/login'
  const response = await fetch(USERS_URL, options);

  if (response.status == 200) {
    let json = await response.json();
    return json;
  } else {
    return [];
  }
}

const getAgents = async() => {
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  //console.log("options", options);
  //config.url + '/awb/api/aq_global/security/login'
  const response = await fetch(AGENTS_URL, options);

  if (response.status == 200) {
    let json = await response.json();
    return json;
  } else {
    return [];
  }
}

const getAvailableAgent = async(agentName) => {
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  const response = await fetch(AGENTS_URL, options);

  if (response.status == 200) {
    let agents = await response.json();
    if(agentName == "any"){
      for(let i = 0; i < agents.length; i++){ // try for a free agent
        var numOfConcurrentJobs = parseInt(agents[i]["numOfConcurrentJobs"]);
        var inprogressJobList = agents[i]["inprogressJobList"].length;
        var inQueueJobList =  agents[i]["inQueueJobList"].length;
        if(inQueueJobList == 0 && (numOfConcurrentJobs > inprogressJobList)){
          console.log("Found free agent:" + agents[i]["agentName"]);
          return agents[i]["agentName"];
        }
      }

      if(agents.length > 0){ // return first agent
        return agents[0]["agentName"];
      }
    } else {
      for(let i = 0; i < agents.length; i++){
        if(agentName.toLowerCase() == agents[i]["agentName"].toLowerCase()){
          return agentName;
        }
      }

      for(let i = 0; i < agents.length; i++){ // return by pattern matching and good one
        var numOfConcurrentJobs = parseInt(agents[i]["numOfConcurrentJobs"]);
        var inprogressJobList = agents[i]["inprogressJobList"].length;
        var inQueueJobList =  agents[i]["inQueueJobList"].length;
        if(agents[i]["agentName"].toLowerCase().includes(agentName.toLowerCase()) && inQueueJobList == 0 && (numOfConcurrentJobs > inprogressJobList)){
          console.log("Found free agent matching name pattern:" + agents[i]["agentName"]);
          return agents[i]["agentName"];
        }
      }
      for(let i = 0; i < agents.length; i++){ // return by pattern matching
        var numOfConcurrentJobs = parseInt(agents[i]["numOfConcurrentJobs"]);
        var inprogressJobList = agents[i]["inprogressJobList"].length;
        var inQueueJobList =  agents[i]["inQueueJobList"].length;
        if(agents[i]["agentName"].toLowerCase().includes(agentName.toLowerCase())){
          console.log("Found free agent matching name pattern:" + agents[i]["agentName"]);
          return agents[i]["agentName"];
        }
      }
    }
  }

  return "";
}

const getJobSummary = async (jobPID) => {
  JOB_SUMMARY_URL = API_ENDPOINT + API_VERSION + "/test-exec/runs/auto/" + jobPID + "?onlySummary=true";
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  var response = await fetch(JOB_SUMMARY_URL, options);

  if (response.status == 200) {
    let json = await response.json();
    return json;
  } else {
    console.error("Failed to Get Job Summary For:", jobPID, " Status:", response.status, ":", response.statusText);
    return;
  }
}



const getConsolidatedReport = async (jobPID) => {
  JOB_CONSIDATED_URL = API_ENDPOINT + API_VERSION + "/test-exec/runs/auto/" + jobPID + "?consolidated=true";
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  var response = await fetch(JOB_CONSIDATED_URL, options);

  if (response.status == 200) {
    let json = await response.json();
    console.log("json is", json);
    return json;
  } else {
    console.error("Failed to Get consolidated report For:", jobPID, " Status:", response.status, ":", response.statusText);
    return;
  }
}

const updateJob = async (jobPID, agentName) => {
  if(agentName == ""){
    return;
  }

  if(agentName && agentName != ""){
    var agentPattern = agentName;
    agentName =  await getAvailableAgent(agentPattern);
    if(agentName == ""){
      console.error("No Agent Found that matches pattern:" + agentPattern);
      throw "No Agent Found that matches pattern:" + agentPattern;
    } else {
      console.log("running on agent " + agentName);
    }
  }


  var jobDetailsUrl = API_ENDPOINT + API_VERSION + "/test-exec/jobs/auto/" + jobPID;

  // https://carnival.accelq.io/awb/api/carnival/CarnivalProject/v1/test-exec/runs/manual/25004/info
  var options = {
    method: 'GET',
    headers:{
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
      'refresh_token': REFRESH_TOKEN,
      'client_id': CLIENT_ID,
      'access_token': ACCESS_TOKEN
    },
    agent:agent
  };

  var response = await fetch(jobDetailsUrl, options);

  if (response.status == 200) {
    let jobDetails = await response.json();
    console.log("jobDeatis",jobDetails);

    var providerHostConfig = jobDetails["hostConfig"]["providerHostConfig"]|| null;
    if(jobDetails["hostConfig"]["providerHostConfig"] && jobDetails["hostConfig"]["providerHostConfig"]["platform"]){
      providerHostConfig = {"platform":{"pid":jobDetails["hostConfig"]["providerHostConfig"]["platform"]["pid"]},
      "browser":{"pid":jobDetails["hostConfig"]["providerHostConfig"]["browser"]["pid"]},
      "os":{"pid":jobDetails["hostConfig"]["providerHostConfig"]["os"]["pid"]}};
    }

    var newHostConfig = {
        "parallelExecution":jobDetails["hostConfig"]["parallelExecution"],
        "headlessTesting":jobDetails["hostConfig"]["headlessTesting"],
        "browserKeepAlive":jobDetails["hostConfig"]["browserKeepAlive"],
        "hostConfigType":jobDetails["hostConfig"]["hostConfigType"],
        "hostBrowser":jobDetails["hostConfig"]["hostBrowser"],
        "hostOS":jobDetails["hostConfig"]["hostOS"],
        "providerHostConfig": providerHostConfig,
        "agentName":agentName,"agentNamePattern":null,"dynamicAgent":false};

    let updatePayload =  {
                          "purpose": jobDetails["purpose"],
                          "captureScreenshot":jobDetails["captureScreenshot"],
                          "startTimestamp":jobDetails["startTimestamp"],
                          "ignoreVerificationComponents":jobDetails["ignoreVerificationComponents"]
                          ,"enableVideo":jobDetails["enableVideo"],"enableSelfHealing":jobDetails["enableSelfHealing"],
                          "enableHtmlCapture":jobDetails["enableHtmlCapture"],"elementWaitTime":jobDetails["elementWaitTime"],
                          "pageLoadWaitTime":jobDetails["pageLoadWaitTime"],"testSuitePid":jobDetails["testSuitePid"],
                          "status":1,"datasetPid":jobDetails["datasetPid"],
                          "hostConfig":newHostConfig,
                          "runConfigParameters":{ pjPid: jobDetails["runConfigParameters"]["pjPid"]},
                          "executionParameters":{},
                          "testJobRecurringInfo":jobDetails["testJobRecurringInfo"] || null,
                          "jobType":jobDetails["jobType"],
                          "parentPid":jobDetails["parentPid"] || null,
                          "notificationEmailIds":jobDetails["notificationEmailIds"],
                          "notifyOnStart":jobDetails["notifyOnStart"],
                          "emailNotificationMode":jobDetails["emailNotificationMode"],
                          "debug":jobDetails["debug"]
                      };


    console.log("updated ", updatePayload);

    var updateOptions = {
      method: 'PUT',
      headers:{
        'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36',
        'refresh_token': REFRESH_TOKEN,
        'client_id': CLIENT_ID,
        'access_token': ACCESS_TOKEN,
        "Content-Type": "application/json"
      },
      agent:agent,
      body:JSON.stringify(updatePayload)
    };

    var jobUpdateUrl = API_ENDPOINT + API_VERSION + '/test-exec/jobs/auto/' + jobPID + '/update';
    var response = await fetch(jobUpdateUrl, updateOptions);
    if (response.status == 204) {
      return "OK";
    } else {
      console.error("Failed to Update AGENT in Job For:", jobPID, " Status:", response.status, ":", response.statusText);
      return;
    }

  } else {
    console.error("Failed to Get Job Summary For:", jobPID, " Status:", response.status, ":", response.statusText);
    return;
  }
}


const setUpBaseURL = (baseURL) => {
    BASE_URL = baseURL.endsWith('/') ? (baseURL): (baseURL + '/');
    API_ENDPOINT =  BASE_URL + "awb/api/" + config.tenant + "/" + config.project + "/";
    LOGIN_URL = BASE_URL + "awb/api/aq_global/security/login";
    USERS_URL = API_ENDPOINT + API_VERSION + "/conf/users";
    AGENTS_URL = API_ENDPOINT + API_VERSION + "/res/agents/relevant-agents";
}

const getResultExternalAccessURL = (jobPid) => {
  return BASE_URL + "#/resultext?tenant=" + config.tenant + "&project=" + config.project + "&resultId=" + jobPid;
}

const sleep = (milliSeconds) => {
  return new Promise(resolve => setTimeout(resolve, milliSeconds));
}

const info = (msg) => {
  console.log("INFO >>> ", msg);
}

const error = (msg) => {
  console.error("ERROR >>> ", msg);
}

const start = async () => {
  var jobSuccess = true;
  config = loadConfig();
  setUpBaseURL(config.url);
  var loginResponse = await doLogin();
  if(!loginResponse){
    error("Failed to Login URL:" + config.url + ", PROJECT:" + config.project + ", TENANT:" + config.tenant + ", USER:" + config.username);
    return false;
  }

  //var users = await getUsers();
  loginResponse = await doLogin();
  if(!loginResponse){
    error("Failed to Login URL:" + config.url + ", PROJECT:" + config.project + ", TENANT:" + config.tenant + ", USER:" + config.username);
    return false;
  }

  var updateResponse = await updateJob(config.jobId, config["agent"]);

  var realJobObj = await triggerJob(config.jobId, getRunParamJsonPayload(config.runParamPayload));
  if(!realJobObj){
    error("Failed to Trigger JOB:" + config.jobId + ", PROJECT:" + config.project + ", TENANT:" + config.tenant + ", USER:" + config.username + ". Please check the CI Job ID");
    return false;
  }
  var attempt = 0;
  var JOB_PICKUP_RETRY_COUNT = 30;
  var jobStatus = "NOT RUN";
  var realJobPid = realJobObj["pid"];
  var passCount = 0, failCount = 0, runningCount = 0, totalCount = 0, notRunCount = 0;
  var jobIDValid = false;
  var consolidated;
  while(true)
  {
    loginResponse = await doLogin();
    if(!loginResponse){
      error("Failed to Login URL:" + config.url + ", PROJECT:" + config.project + ", TENANT:" + config.tenant + ", USER:" + config.username);
      return false;
    }
    var summaryObj = await getJobSummary(realJobPid);


    if(!summaryObj){
      if(!jobIDValid){
        error("Failed to run job:" + config.jobId + ". Please check the CI Job ID");
      } else {
        error("GET JOB SUMMARY FAILED for:" + realJobPid);
      }
      return false;
    }
    if(summaryObj["cause"]) {
        error(summaryObj["cause"]);
        return false;
    }
    if(summaryObj["summary"] != null) {
        summaryObj = summaryObj["summary"];
    }
    jobIDValid = true;
    var passCount = summaryObj["pass"];
    var failCount = summaryObj["fail"];
    var notRunCount = summaryObj["notRun"];
    jobStatus = summaryObj["status"].toUpperCase();

    console.log(summaryObj);

    info(jobStatus + " Accelq Run:" + realJobPid);
    info("Status: " + summaryObj["status"]);
    info("Pass: " + passCount);
    info("Fail: " + failCount);
    info("Not Run: " + notRunCount);
    info("");

    if(jobStatus == "SCHEDULED"){
        attempt = attempt + 1;
    }
    if(attempt == JOB_PICKUP_RETRY_COUNT) {
        error("No agent available to pickup the job");
        jobSuccess = false;
        break;
    }
    if((jobStatus == "COMPLETED") || (jobStatus == "ABORTED") || (jobStatus == "FAILED TO START")){
      break;
    }

    if((jobStatus == "ABORTED") || (jobStatus == "FAILED TO START")){
      jobSuccess = false;
      break;
    }
    await sleep(30 * 1000);
  }

  // consolidated = await getConsolidatedReport(realJobPid);
  console.log("cons", summaryObj);
  generateJUnitReport(summaryObj);
  info("######### RESULTS SUMMARY ##########");
  info("Status: " + jobStatus);
  info("Pass: " + passCount);
  info("Fail: " + failCount);
  info("Not Run: " + notRunCount);
  info("");
  info("#####################################");
  var resultAccessURL = getResultExternalAccessURL(realJobPid);
  console.log(" For Full report Please Visit:", resultAccessURL);
  if(failCount > 0
          || (jobStatus == "ABORTED")
          || (jobStatus == "FAILED TO START")) {
      error("Run Failed");
      jobSuccess = false;
  }

  return jobSuccess;
}

// start the flow
try {
  start().then((success) => {
    if(!success){
      error("Build Failed");
      process.exit(1);
    }
    process.exit(0);
  });
} catch (e) {
  console.log(e);
  process.exit(1);
}
